import functools
import os
import ssl
import tempfile

import grpclib.exceptions
from grpclib.client import Channel
from grpclib.protocol import H2Protocol
import _ssl
import asyncio

from typing import List, Tuple, cast, TypeVar, Callable, Any, AsyncGenerator

from cryptography.x509 import CertificateSigningRequest, Certificate
from cryptography.hazmat.primitives.asymmetric.rsa import (
    RSAPrivateKey,
)

# FIXME: the next line should be fixed when Fedora has
# protoc 3.19.0 or later, and the protobufs need to be recompiled
# when that happens.  Not just the hassmpris protos, also the
# cakes ones.
os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"

from google.protobuf.empty_pb2 import Empty  # noqa: E402
from hassmpris.proto import mpris_grpc  # noqa: E402

import cakes  # noqa: E402
import blindecdh  # noqa: E402

from hassmpris.proto import mpris_pb2  # noqa: E402
import hassmpris.certs as certs  # noqa: E402


__version__ = "0.0.19"
__SPEC_URL = (
    "https://specifications.freedesktop.org/"
    "mpris-spec/2.2/Player_Interface.html#methods"
)

DEFAULT_TIMEOUT = 15.0


Ignored = cakes.Ignored
Rejected = cakes.Rejected
CannotDecrypt = cakes.CannotDecrypt


class ClientException(Exception):
    """
    The base class for all HASS MPRIS client exceptions.
    """


class CannotConnect(ClientException):
    """
    The remote server is not running or refuses connections.
    """

    def __str__(self) -> str:
        f = "Server is not running or refuses connections: %s"
        return f % (self.args[0],)


class Unauthenticated(ClientException):
    """
    The server has not authenticated us.
    """

    def __str__(self) -> str:
        f = "Client is not authenticated: %s"
        return f % (self.args[0],)


class Disconnected(ClientException):
    """
    The server is now gone.
    """

    def __str__(self) -> str:
        f = "Server gone: %s"
        return f % (self.args[0],)


class Timeout(ClientException):
    """
    The connection to the server timed out.
    """

    def __str__(self) -> str:
        f = "Server timed out: %s"
        return f % (self.args[0],)


StubFunc = TypeVar("StubFunc", bound=Callable[..., Any])


def normalize_connection_errors(f: StubFunc) -> StubFunc:
    """
    Normalizes connection errors for easier handling.
    """

    @functools.wraps(f)
    async def inner(*args: Tuple[Any]) -> Any:
        try:
            return await f(*args)
        except ssl.SSLCertVerificationError as e:
            raise Unauthenticated(e)
        except ConnectionRefusedError as e:
            raise CannotConnect(e)
        except OSError as e:
            raise CannotConnect(e)
        except grpclib.exceptions.StreamTerminatedError as e:
            raise Disconnected(e)
        except asyncio.exceptions.TimeoutError as e:
            raise Timeout(e)

    return cast(StubFunc, inner)


def normalize_connection_errors_iterable(f: StubFunc) -> StubFunc:
    """
    Normalizes connection errors in async generators for easier handling.
    """

    @functools.wraps(f)
    async def inner(*args: Tuple[Any]) -> Any:
        try:
            async for x in f(*args):
                yield x
        except ssl.SSLCertVerificationError as e:
            raise Unauthenticated(e)
        except ConnectionRefusedError as e:
            raise CannotConnect(e)
        except OSError as e:
            raise CannotConnect(e)
        except asyncio.exceptions.CancelledError as e:
            raise Disconnected(e)
        except grpclib.exceptions.StreamTerminatedError as e:
            raise Disconnected(e)
        except asyncio.exceptions.TimeoutError as e:
            raise Timeout(e)

    return cast(StubFunc, inner)


class AsyncCAKESClient(object):
    """
    The CAKES client class to securely pair an MPRIS client to the
    MPRIS desktop agent.

    This is a wrapper around cakes.client.AsyncCAKESClient that brings
    its own channel -- so you don't have to provide one.

    See file cli.py in the same folder as the file containing this class
    for a sample minimal client you can use in your own projects.
    """

    def __init__(
        self,
        host: str,
        port: int,
        csr: CertificateSigningRequest,
    ):
        """
        Initialize the CAKES client.

        Parameters:
          host: the host name to connect to
          port: the CAKES server port (customarily it is port 40052)
          csr:  a CertificateSigningRequest you provide, in order for the
                server to issue a valid certificate after successful pairing.
        """
        self.channel = grpclib.client.Channel(host, port)
        self.client = cakes.AsyncCAKESClient(
            self.channel,
            csr,
        )
        self.ecdh: blindecdh.CompletedECDH | None = None

    def __del__(self) -> None:
        delattr(self, "client")
        self.channel.close()
        delattr(self, "channel")

    @normalize_connection_errors
    async def obtain_verifier(self) -> blindecdh.CompletedECDH:
        """
        Obtains the verifier with the derived_key attribute.

        Compare this derived_key with the counterparty's derived_key.

        If they do not match, DO NOT call obtain_certificate() --
        your communication is compromised.

        Refer to cakes.AsyncCAKESClient for more documentation.
        """
        self.ecdh = await self.client.obtain_verifier()
        return self.ecdh

    @normalize_connection_errors
    async def obtain_certificate(
        self,
    ) -> Tuple[Certificate, List[Certificate]]:  # noqa:E501
        """
        Obtains the signed client certificate and the trust chain.

        Only call when obtain_verifier()'s result has been verified
        to match on both sides.

        Refer to cakes.AsyncCAKESClient for more documentation.
        """
        assert self.ecdh, "did not run obtain_verifier"
        return await self.client.obtain_certificate(self.ecdh)


class MPRISChannel(Channel):
    """
    A secure gRPC channel that overrides the hostname to an expected value.

    All HASS MPRIS servers use hostname 'hassmpris', because they are not
    bound to the global DNS system.  gRPC does not permit by default to
    override the server hostname, so we must add this glue to make this
    happen.
    """

    def __init__(
        self,
        host: str,
        port: int,
        client_cert: Certificate,
        client_key: RSAPrivateKey,
        trust_chain: List[Certificate],
    ):
        self._client_cert = client_cert
        self._client_key = client_key
        self._trust_chain = trust_chain
        Channel.__init__(self, host, port, ssl=True)

    def _get_default_ssl_context(self) -> "_ssl.SSLContext":
        with tempfile.TemporaryDirectory() as d:
            certs.save_client_certs_and_trust_chain(
                d,
                self._client_cert,
                self._client_key,
                self._trust_chain,
            )
            c = os.path.join(d, "client.crt")
            k = os.path.join(d, "client.key")
            t = os.path.join(d, "client.trust.pem")
            ctx = ssl.create_default_context(
                purpose=ssl.Purpose.SERVER_AUTH,
            )
            ctx.load_cert_chain(c, k)
            ctx.load_verify_locations(cafile=t)
            ctx.check_hostname = True
            ciphers = "ECDHE+AESGCM:ECDHE+CHACHA20:DHE+AESGCM:DHE+CHACHA20"
            ctx.set_ciphers(ciphers)
            ctx.set_alpn_protocols(["h2"])
        return ctx

    async def _create_connection(self) -> H2Protocol:
        _, protocol = await self._loop.create_connection(
            self._protocol_factory,
            self._host,
            self._port,
            ssl=self._ssl,
            server_hostname="hassmpris",
        )
        return protocol


class AsyncMPRISClient(object):
    """
    The MPRIS client class to govern the MPRIS desktop agent remotely.

    See file cli.py in the same folder as the file containing this class
    for a sample minimal client you can use in your own projects.
    """

    def __init__(
        self,
        host: str,
        port: int,
        client_cert: Certificate,
        client_key: RSAPrivateKey,
        trust_chain: List[Certificate],
    ) -> None:
        """
        Initialize the client.

        You must already have the client_cert, client_key and trust_chain
        values.  If you don't, you must use the AsyncCAKESClient class to
        obtain it.

        Parameters:
          host: the host name to connect to
          port: the port (customarily, it is port 40051)
          client_cert, client_key, trust_chain: cryptographic material
                associated to your MPRIS agent.
        """
        self.host = host
        self.channel = MPRISChannel(
            host,
            port,
            client_cert,
            client_key,
            trust_chain,
        )
        self.stub = mpris_grpc.MPRISStub(channel=self.channel)

    def __del__(self) -> None:
        if hasattr(self, "stub"):
            delattr(self, "stub")
        if hasattr(self, "channel"):
            self.channel.close()
            delattr(self, "channel")

    async def close(self) -> None:
        """
        Clean up the client.

        Any concomitantly running stream_updates() will raise an exception and
        terminate execution.

        This method is called when the garbage collector disposes of the
        client, so you want to keep a reference to the client alive if you want
        this object to continue connected to the server.
        """
        self.__del__()

    @normalize_connection_errors
    async def ping(self) -> None:
        """
        Ping the server, verifying cryptography is working.

        A number of exceptions may be raised.  See the code for the function
        normalize_connection_errors to discover the most common exceptions your
        code will have to deal with.
        """
        await self.stub.Ping(Empty(), timeout=DEFAULT_TIMEOUT)

    @normalize_connection_errors_iterable
    async def stream_updates(
        self,
    ) -> AsyncGenerator[mpris_pb2.MPRISUpdateReply, None]:
        """
        Generate a stream of MPRISUpdateReply, yielding them asynchronously
        to the caller.

        This is an async generator.

        A number of exceptions may be raised.  See the code for the function
        normalize_connection_errors_iterable to discover the most common
        exceptions your code will have to deal with.
        """
        async with self.stub.Updates.open() as stream:
            await stream.send_message(mpris_pb2.MPRISUpdateRequest(), end=True)
            async for message in stream:
                yield message

    @normalize_connection_errors
    async def change_player_status(
        self,
        player_id: str,
        playback_status: int,
    ) -> mpris_pb2.ChangePlayerStatusReply:
        """
        Change player status to one of the states enumerated in
        mpris_pb2.ChangePlayerStatusRequest.PlaybackStatus.

        You want to use the direct methods pause, play or stop.
        """
        return await self.stub.ChangePlayerStatus(
            mpris_pb2.ChangePlayerStatusRequest(
                player_id=player_id,
                status=playback_status,
            ),
            timeout=DEFAULT_TIMEOUT,
        )

    async def pause(
        self,
        player_id: str,
    ) -> mpris_pb2.ChangePlayerStatusReply:
        """
        Tell the server to pause playback of one player.

        Parameters:
          player_id: a player ID as per one of the MPRISUpdateRequest received.

        A number of exceptions may be raised.  See the code for the function
        normalize_connection_errors to discover the most common exceptions your
        code will have to deal with.
        """
        pbstatus = mpris_pb2.ChangePlayerStatusRequest.PlaybackStatus
        return await self.change_player_status(player_id, pbstatus.PAUSED)

    async def play(
        self,
        player_id: str,
    ) -> mpris_pb2.ChangePlayerStatusReply:
        """
        Tell the server to begin playback in one player.

        Parameters:
          player_id: a player ID as per one of the MPRISUpdateRequest received.

        A number of exceptions may be raised.  See the code for the function
        normalize_connection_errors to discover the most common exceptions your
        code will have to deal with.
        """
        pbstatus = mpris_pb2.ChangePlayerStatusRequest.PlaybackStatus
        return await self.change_player_status(player_id, pbstatus.PLAYING)

    async def stop(
        self,
        player_id: str,
    ) -> mpris_pb2.ChangePlayerStatusReply:
        """
        Tell the server to stop playback of one player.

        Parameters:
          player_id: a player ID as per one of the MPRISUpdateRequest received.

        A number of exceptions may be raised.  See the code for the function
        normalize_connection_errors to discover the most common exceptions your
        code will have to deal with.
        """
        pbstatus = mpris_pb2.ChangePlayerStatusRequest.PlaybackStatus
        return await self.change_player_status(player_id, pbstatus.STOPPED)

    async def previous(
        self,
        player_id: str,
    ) -> mpris_pb2.PreviousReply:
        """
        Tells the server to skip one track backward in one player.

        Parameters:
          player_id: a player ID as per one of the MPRISUpdateRequest received.

        A number of exceptions may be raised.  See the code for the function
        normalize_connection_errors to discover the most common exceptions your
        code will have to deal with.
        """
        m = mpris_pb2.PreviousRequest(player_id=player_id)
        return await self.stub.Previous(m)

    async def next(
        self,
        player_id: str,
    ) -> mpris_pb2.PreviousReply:
        """
        Tells the server to skip one track forward in one player.

        Parameters:
          player_id: a player ID as per one of the MPRISUpdateRequest received.

        A number of exceptions may be raised.  See the code for the function
        normalize_connection_errors to discover the most common exceptions your
        code will have to deal with.
        """
        m = mpris_pb2.NextRequest(player_id=player_id)
        return await self.stub.Next(m)

    async def seek(
        self,
        player_id: str,
        offset: float,
    ) -> mpris_pb2.SeekReply:
        """
        Tells the server to seek within the current-playing track of a player.

        Parameters:
          player_id: a player ID as per one of the MPRISUpdateRequest received.
          offset: a positive or negative float indicating how many seconds to
            go forward or backward.

        A number of exceptions may be raised.  See the code for the function
        normalize_connection_errors to discover the most common exceptions your
        code will have to deal with.
        """
        m = mpris_pb2.SeekRequest(player_id=player_id, offset=offset)
        return await self.stub.Seek(m)

    async def set_position(
        self,
        player_id: str,
        track_id: str,
        position: float,
    ) -> mpris_pb2.SetPositionReply:
        f"""
        Tells the server to play the track ID from the currently playing track.

        Parameters:
          player_id: a player ID as per one of the MPRISUpdateRequest received.
          track_id: the string ID "mpris:trackid" sent by the player in the
            metadata update payload.  If this does not match the current track,
            the set position command will be ignored as stale.
          position: an absolute zero or positive float indicating how many
            seconds to go into the track from its beginning.

        See {__SPEC_URL}
        for more information.

        A number of exceptions may be raised.  See the code for the function
        normalize_connection_errors to discover the most common exceptions your
        code will have to deal with.
        """
        m = mpris_pb2.SetPositionRequest(
            player_id=player_id, track_id=track_id, position=position
        )
        return await self.stub.SetPosition(m)



import errno
import os
import shlex
import sys

import asyncio


import shortauthstrings  # noqa: E402

# FIXME: the next line should be fixed when Fedora has
# protoc 3.19.0 or later, and the protobufs need to be recompiled
# when that happens.  Not just the hassmpris protos, also the
# cakes ones.
os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"

from hassmpris_client import (  # noqa: E402
    AsyncMPRISClient,
    AsyncCAKESClient,
    Rejected,
    Unauthenticated,
)
import blindecdh  # noqa: E402

from hassmpris.proto import mpris_pb2  # noqa: E402
import hassmpris.certs as certs  # noqa: E402

from hassmpris import config  # noqa: E402


def accept_ecdh_via_console(
    peer: str,
    complete: blindecdh.CompletedECDH,
) -> bool:
    msg = f"""
A notification has appeared on the remote computer {peer}.
Please use it to verify that the key matches this computer's.
""".strip()
    print(msg)
    print(
        "The key on this side appears to be %s"
        % shortauthstrings.emoji(
            complete.derived_key,
            6,
        )
    )
    print("Accept?  [Y/N then ENTER]")
    line = sys.stdin.readline()
    result = line.lower().startswith("y")
    return result


async def repl(stub: AsyncMPRISClient, known_players: list[str]) -> None:
    print(
        "When you open an MPRIS-compatible player, you will see its name scroll onscreen."  # noqa: E501
    )

    def help() -> None:
        print("Commands:")
        print(
            "* play [optionally player name]      \n"
            "  plays media on the player            "
        )
        print(
            "* pause [optionally player name]     \n"
            "  pauses media on the player           "
        )
        print(
            "* stop [optionally player name]      \n"
            "  stops media on the player            "
        )
        print(
            "* prev [optionally player name]      \n"
            "  skip to previous track               "
        )
        print(
            "* next [optionally player name]      \n"
            "  skip to next track                   "
        )
        print(
            "* seek <pos> [optionally player name]\n"
            "  seek player to pos (in seconds)      "
        )
        print(
            "* empty line                         \n"
            "  shows this help message              "
        )
        print(
            "* Ctrl+D / close stdin               \n"
            "  exits this client application        "
        )

    help()
    loop = asyncio.get_running_loop()
    fd = sys.stdin.fileno()
    while True:
        future = asyncio.Future()  # type: ignore
        loop.add_reader(fd, future.set_result, None)
        future.add_done_callback(lambda f: loop.remove_reader(fd))
        line = await future
        line = sys.stdin.readline()
        if not line:
            return
        s = line.strip()
        if not s:
            help()
            continue

        tokens = shlex.split(s)
        cmd, parms = tokens[0], tokens[1:]

        if cmd == "seek":
            if len(parms) == 1:
                if not known_players:
                    print(
                        "There is no last player to commandeer.",
                        file=sys.stderr,
                    )
                    continue
                offset, player = float(parms[0]), known_players[-1]
            elif len(parms) > 1:
                offset, player = float(parms[0]), parms[1]
            else:
                print(
                    "You must specify an offset to seek to",
                    file=sys.stderr,
                )
                continue
        else:
            if len(parms) == 0:
                if not known_players:
                    print(
                        "There is no last player to commandeer.",
                        file=sys.stderr,
                    )
                    continue
                player = known_players[-1]
            else:
                player = parms[0]

        try:
            if cmd == "pause":
                await stub.pause(player)
            elif cmd == "play":
                await stub.play(player)
            elif cmd == "stop":
                await stub.stop(player)
            elif cmd == "prev":
                await stub.previous(player)
            elif cmd == "next":
                await stub.next(player)
            elif cmd == "seek":
                await stub.seek(player, offset)
        except Exception as e:
            print(
                "Cannot commandeer player %s because of %s %s"
                % (
                    player,
                    type(e),
                    e,
                ),
                file=sys.stderr,
            )


async def print_updates(
    mprisclient: AsyncMPRISClient,
    players: list[str],
) -> None:
    # FIXME: the server is not sending me the status of the player
    # when it initially streams the players it knows about.
    async for update in mprisclient.stream_updates():
        print(update)
        if update.HasField("player"):
            if update.player.status == mpris_pb2.PlayerStatus.GONE:
                while update.player.player_id in players:
                    players.remove(update.player.player_id)
            elif update.player.player_id not in players:
                players.append(update.player.player_id)


def usage() -> str:
    prog = sys.argv[0]
    usage_str = f"""
usage: {prog} <server> [ping]

If ping is specified as the second parameter, then the program will simply
attempt to ping the server and exit immediately if successful.

If ping is not specified, you get a rudimentary remote control.
""".strip()
    return usage_str


async def async_main() -> int:
    if not sys.argv[1:]:
        print(usage())
        return os.EX_USAGE
    server = sys.argv[1]
    action = sys.argv[2] if sys.argv[2:] else None

    try:
        (
            client_cert,
            client_key,
            trust_chain,
        ) = certs.load_client_certs_and_trust_chain(config.folder())
        cakes_needed = False
    except FileNotFoundError:
        cakes_needed = True

    if cakes_needed:
        client_csr, client_key = certs.create_and_load_client_key_and_csr(
            config.folder()
        )
        cakesclient = AsyncCAKESClient(
            server,
            40052,
            client_csr,
        )
        try:
            ecdh = await cakesclient.obtain_verifier()
        except Rejected as e:
            print("Not authorized: %s" % e)
            return errno.EACCES

        result = accept_ecdh_via_console(server, ecdh)
        if not result:
            print("Locally rejected.")
            return errno.EACCES

        try:
            client_cert, trust_chain = await cakesclient.obtain_certificate()
        except Rejected as e:
            print("Not authorized: %s" % e)
            return errno.EACCES

        certs.save_client_certs_and_trust_chain(
            config.folder(),
            client_cert,
            client_key,
            trust_chain,
        )

    mprisclient = AsyncMPRISClient(
        server,
        40051,
        client_cert,
        client_key,
        trust_chain,
    )
    try:
        if action == "ping":
            await mprisclient.ping()
            print("Successfully pinged the server.")
        else:
            players: list[str] = []

            replfuture = asyncio.create_task(
                repl(mprisclient, players),
            )
            updatesfuture = asyncio.create_task(
                print_updates(mprisclient, players),
            )

            try:
                done, pending = await asyncio.wait(
                    [
                        replfuture,
                        updatesfuture,
                    ],
                    return_when=asyncio.FIRST_COMPLETED,
                )
                done.pop().result()
            except Exception:
                os.close(0)
                raise

    except Unauthenticated:
        print("Server has reset its certificate store.")
        print("Remove client files in ~/.config/hassmpris to reauthenticate.")
        return errno.EACCES

    return 0


def main() -> None:
    sys.exit(asyncio.run(async_main()))


if __name__ == "__main__":
    main()
